from ..data.DataWrapper import DataWrapper
import numpy as np
import copy
import torch

def configure_division(base_dataset, repeats, n_openness=None, seed=None, min_known_classes=2):
        """
        Method for obtaining configurations for OSR model evaluation using Holdout protocol (both KKC and UUC from single dataset).

        :type base_dataset: VisionDataset
        :param base_dataset: Base dataset
        
        :type repeats: int
        :param repeats: Number of randol selections of classes for single openness (KKC/UUC class cardinality)
        
        :type n_openness: int
        :param n_openness: Number of KKC/UUC class cardinality to generate. If None will return all possible configurations.
        
        :type seed: int
        :param seed: Random state
        
        :type min_known_classes: int
        :param min_known_classes: Minimum number of known classes
        
        :rtype: List
        :returns: Lit of dataset configurations -- each containing sets of KKC and UUC -- and their Openness
        """
        # Set random generator with manual seed
        trng = torch.random.manual_seed(seed)
        
        # Get number of classes from base dataset
        n_classes = (base_dataset._n_classes())

        # Tables for all possible numbers of KKC and UUC classes
        kkc, uuc = torch.triu_indices(n_classes-1, n_classes-1)
        kkc += 1
        uuc = n_classes - 1 - uuc
        
        # Remove all the entries that have less known classes than given minimum (default=2)
        kkc, uuc = kkc[kkc>=min_known_classes], uuc[kkc>=min_known_classes]

        # Calculate openness for possible KKC and UUC
        openness_n_classes = torch.stack((kkc,uuc)).T        
        openness = 1 - torch.sqrt((2*kkc)/((2*kkc)+uuc))

        # Randomly select n_openness configurations from possible KKC, UUC and openness if n_openness is not None
        if n_openness is not None:
                # Assign higher selection probability to sets with larger number of classes
                p = openness_n_classes.sum(1)
                p = p / p.sum()
                rand_indexes = p.multinomial(num_samples=n_openness, replacement=False, generator=trng)

                # Get selected openness and KKC/UUC configurations
                op_choice = openness[rand_indexes]
                op_n_classes_choice = openness_n_classes[rand_indexes]
        else:
                # Id n_openness not specified return all possible configurations
                op_choice = openness
                op_n_classes_choice = openness_n_classes

        # For selected konfigurations randomly assign classes
        repeats_config =[]
        for kkc_n, uuc_n in op_n_classes_choice:    
                for r in range(repeats):
                        all_classes = torch.ones(n_classes)
                        kkc = all_classes.multinomial(num_samples=kkc_n, replacement=False, generator=trng)
                        all_classes[kkc] = 0
                        uuc = all_classes.multinomial(num_samples=uuc_n, replacement=False, generator=trng)
                        
                        repeats_config.append((kkc, uuc))
 
        return repeats_config, op_choice

def get_train_test(base_dataset, kkc_indexes, uuc_indexes, root, tunning, fold, seed=1410, n_folds=5):
        """
        Method for obtaining Cross-validation folds using Holdout protocol (both KKC and UUC from single dataset).

        :type base_dataset: VisionDataset
        :param base_dataset: Base dataset
        
        :type kkc_indexes: List
        :param kkc_indexes: List of labels constituting Known Classes        
        
        :type uuc_indexes: List
        :param uuc_indexes: List of labels constituting Unknown Classes
        
        :type root: string
        :param root: Datasets folder
        
        :type tunning: boolean
        :param tunning: Flag. If True will split 10% of data for tunning, otherwise will split 90% of data.
                
        :type fold: int
        :param fold: Fold index
        
        :type n_folds: int
        :param n_folds: Number of folds
        
        :type seed: int
        :param seed: Random state
        
        :rtype: List
        :returns: Train dataset, Test dataset
        """
        
        trng = torch.random.manual_seed(seed)
        
        classes = np.concatenate((kkc_indexes, uuc_indexes))

        # Generate dataset
        data = DataWrapper(root=root,
                       base_dataset=base_dataset,
                       indexes='all',
                       get_classes=classes,
                       known_classes=kkc_indexes,
                       return_only_known=False,
                       onehot=True,
                       onehot_num_classes=len(kkc_indexes)+1
                       )
        
        labels = data.original_targets.unique()
        
        # Define portion of data that will be considered tunning set
        proportion = .1
        
        # First tuning, later validation
        division_probability = torch.ones(n_folds*2) * (1-proportion) / n_folds
        division_probability[:n_folds] = proportion / n_folds
        
        division_mask = torch.zeros_like(data.targets)
        
        for label in labels:
                class_mask = data.original_targets == label
                division_mask[class_mask] = division_probability.multinomial(class_mask.sum(0), replacement=True, generator=trng)                

        if tunning:
                test_mask = division_mask == fold
                train_mask = division_mask < n_folds
                
                train_mask = train_mask * ~test_mask
        else:
                test_mask = division_mask == (fold+n_folds)
                train_mask = division_mask >= n_folds
                
                train_mask = train_mask * ~test_mask
        
        train_indexes = torch.where(train_mask)[0]
        test_indexes = torch.where(test_mask)[0]
        
        # REMOVE UUC FROM TRAINING
        uuc_idx = torch.argwhere(data.targets == len(kkc_indexes))
        train_indexes = np.array([i for i in train_indexes if i not in uuc_idx])

        data_test = copy.deepcopy(data)
        data_train = copy.deepcopy(data)

        data_test.reindex(test_indexes)
        data_train.reindex(train_indexes)  
        
        return data_train, data_test




import numpy as np
import torch

def configure_oneclass_division(base_dataset, repeats, n_openness=None, seed=None):
        """
        Method for obtaining configurations for OSR model evaluation using Holdout protocol (both KKC and UUC from single dataset) for One-class classification. Set of KKC always contains a single class.

        :type base_dataset: VisionDataset
        :param base_dataset: Base dataset
        
        :type repeats: int
        :param repeats: Number of randol selections of classes for single openness (KKC/UUC class cardinality)
        
        :type n_openness: int
        :param n_openness: Number of KKC/UUC class cardinality to generate. In None will return all possible configurations.
        
        :type seed: int
        :param seed: Random state
        
        :rtype: List
        :returns: List of dataset configurations -- each containing sets of KKC and UUC -- and their Openness
        """
        
        # Set random generator with manual seed
        trng = torch.random.manual_seed(seed)
        
        # Get number of classes from base dataset
        n_classes = (base_dataset._n_classes())

        # Tables for all possible numbers of KKC and UUC classes
        openness_n_classes = []
        openness = []
        
        for j in range(1,n_classes-1):
                openness_n_classes.append([1,j])
                openness.append(1 - np.sqrt((2)/(2+j)))
                
        openness = torch.tensor(openness)
        openness_n_classes = torch.tensor(openness_n_classes)
        
        # Jeżeli mamy n_openness losujemy n konfiguracji
        if n_openness is not None:
                # Faworyzujemy te podzbiory, w których sumaryczna liczba klas jest duża
                p = openness_n_classes.sum(1)
                p = p / p.sum()
                rand_indexes = p.multinomial(num_samples=n_openness, replacement=False, generator=trng)

                # Pobranie wylosowanych liczności i openness
                op_choice = openness[rand_indexes]
                op_n_classes_choice = openness_n_classes[rand_indexes]
        else:
                # W innym wypadku pobieramy wszystkie
                op_choice = openness
                op_n_classes_choice = openness_n_classes

        #Teraz dla wylosowanych liczności będziemy losować klasy
        repeats_config =[]

        for kkc_n, uuc_n in op_n_classes_choice:    
                #Dla liczby powtórzeń losujemy indeksy kkc i uuc
                for r in range(repeats):
                        all_classes = torch.ones(n_classes)
                        kkc = all_classes.multinomial(num_samples=kkc_n, replacement=False, generator=trng)
                        all_classes[kkc] = 0
                        uuc = all_classes.multinomial(num_samples=uuc_n, replacement=False, generator=trng)
                        
                        repeats_config.append((kkc, uuc))
 
        return repeats_config, op_choice




"""Basic tests."""

import numpy as np
import torchosr as osr
import torch
from torchosr.data.base_datasets import SVHN_base
from torchosr.data import DataWrapper
from torchosr.utils.base import grayscale_transform
from torchvision import transforms
from torch.utils.data import DataLoader

def get_data(known):
    t_svhn = transforms.Compose([
        grayscale_transform(),
        transforms.Resize(28),
        transforms.ToTensor()])

    base_data = SVHN_base(root='data',transform=t_svhn, download=True)
    train_data = DataWrapper(root='data', 
                    base_dataset=base_data,
                    indexes=np.arange(0,6000),
                    get_classes=np.arange(10),
                    known_classes=known,
                    return_only_known=True,
                    onehot=True,
                    onehot_num_classes=len(known)+1)

    test_data = DataWrapper(root='data', 
                    base_dataset=base_data,
                    indexes=np.arange(6000,7000),
                    get_classes=np.arange(10),
                    known_classes=known,
                    return_only_known=False,
                    onehot=True,
                    onehot_num_classes=len(known)+1)
    
    return train_data, test_data

def test_NoiseSoftmax():
    # Define parameters
    learning_rate = 1e-3
    batch_size = 128
    epochs = 2

    # Select data
    known = [0,1]
    train_data, test_data = get_data(known)
    
    # Prepare data loaders
    train_data_loader = DataLoader(train_data, batch_size=batch_size, shuffle=True)
    test_data_loader = DataLoader(test_data, batch_size=batch_size, shuffle=True)
    
    # Initialize model
    model_openmax = osr.models.NoiseSoftmax(n_known=len(known))
    
    # Initialize loss function
    loss_fn = torch.nn.CrossEntropyLoss()
    
    # Initialize optimizer
    optimizer_openmax = torch.optim.SGD(model_openmax.parameters(), lr=learning_rate)
    
    for t in range(epochs):
        print(f"Epoch {t+1}\n-------------------------------")
        inner_score, outer_score, hp_score, overall_score = model_openmax.test(test_data_loader, loss_fn)
        model_openmax.train(train_data_loader, loss_fn, optimizer_openmax)
        
def test_OverlaySoftmax():
    # Define parameters
    learning_rate = 1e-3
    batch_size = 128
    epochs = 2

    # Select data
    known = [0,1]
    train_data, test_data = get_data(known)
    
    # Prepare data loaders
    train_data_loader = DataLoader(train_data, batch_size=batch_size, shuffle=True)
    test_data_loader = DataLoader(test_data, batch_size=batch_size, shuffle=True)
    
    # Initialize model
    model_openmax = osr.models.OverlaySoftmax(n_known=len(known))
    
    # Initialize loss function
    loss_fn = torch.nn.CrossEntropyLoss()
    
    # Initialize optimizer
    optimizer_openmax = torch.optim.SGD(model_openmax.parameters(), lr=learning_rate)
    
    for t in range(epochs):
        print(f"Epoch {t+1}\n-------------------------------")
        inner_score, outer_score, hp_score, overall_score = model_openmax.test(test_data_loader, loss_fn)
        model_openmax.train(train_data_loader, loss_fn, optimizer_openmax)
    
def test_Openmax():
    # Define parameters
    learning_rate = 1e-3
    batch_size = 128
    epochs = 2
    epsilon = .5

    # Select data
    known = [0,1]
    train_data, test_data = get_data(known)
    
    # Prepare data loaders
    train_data_loader = DataLoader(train_data, batch_size=batch_size, shuffle=True)
    test_data_loader = DataLoader(test_data, batch_size=batch_size, shuffle=True)
    
    # Initialize model
    model_openmax = osr.models.Openmax(n_known=len(known), epsilon=epsilon)
    
    # Initialize loss function
    loss_fn = torch.nn.CrossEntropyLoss()
    
    # Initialize optimizer
    optimizer_openmax = torch.optim.SGD(model_openmax.parameters(), lr=learning_rate)
    
    for t in range(epochs):
        print(f"Epoch {t+1}\n-------------------------------")
        inner_score, outer_score, hp_score, overall_score = model_openmax.test(test_data_loader, loss_fn)
        model_openmax.train(train_data_loader, loss_fn, optimizer_openmax)

def test_Softmax():
    # Define parameters
    learning_rate = 1e-3
    batch_size = 128
    epochs = 2
    epsilon = .5

    # Select data
    known = [0,1]
    train_data, test_data = get_data(known)
    
    # Prepare data loaders
    train_data_loader = DataLoader(train_data, batch_size=batch_size, shuffle=True)
    test_data_loader = DataLoader(test_data, batch_size=batch_size, shuffle=True)
    
    # Initialize model
    model_openmax = osr.models.TSoftmax(n_known=len(known), epsilon=epsilon)
    
    # Initialize loss function
    loss_fn = torch.nn.CrossEntropyLoss()
    
    # Initialize optimizer
    optimizer_openmax = torch.optim.SGD(model_openmax.parameters(), lr=learning_rate)
    
    for t in range(epochs):
        print(f"Epoch {t+1}\n-------------------------------")
        inner_score, outer_score, hp_score, overall_score = model_openmax.test(test_data_loader, loss_fn)
        model_openmax.train(train_data_loader, loss_fn, optimizer_openmax)


"""Data tests."""

import torch
import torchosr as to
from torchvision import transforms

# TRANSFORMS
t_mnist = transforms.Compose([
        transforms.Resize(28),
        transforms.ToTensor()])

t_omni = transforms.Compose([
        transforms.Resize(28),
        transforms.ToTensor(),
        to.utils.inverse_transform()])

t_svhn = transforms.Compose([
        transforms.Resize(32),
        transforms.ToTensor()])

t_cifar = transforms.Compose([
        transforms.Resize(32),
        transforms.ToTensor()])

root = 'data'
n_openness = 2
repeats = 2
n_splits = 5
seed = 43


def test_MNIST_Omni_config():
    base = to.data.base_datasets.MNIST_base(root=root, download=True, transform=t_mnist)
    out = to.data.base_datasets.Omniglot_base(root=root, download=True, transform=t_omni)

    config = to.data.configure_division_outlier(base, out, n_openness, repeats, seed)
    
    for config_i, (kkc, uuc) in enumerate(config[0]):
            train, test = to.data.get_train_test_outlier(base, out, kkc, uuc, 
                                                         root=root, 
                                                         tunning=True, 
                                                         fold=1, 
                                                         seed=seed)
            
            train_data = [train.__getitem__(i)[0] for i in range(len(train))]
            test_data = [test.__getitem__(i)[0] for i in range(len(test))]
            
            for trd_i, trd in enumerate(train_data):
                    for ted_i, ted in enumerate(test_data):
                            if torch.equal(trd, ted):
                                    exit()


def test_MNIST_config():
    base = to.data.base_datasets.MNIST_base(root=root, download=True, transform=t_mnist)
    config = to.data.configure_division(base, n_openness, repeats, seed)

    for config_i, (kkc, uuc) in enumerate(config[0]):
            train, test = to.data.get_train_test(base, kkc, uuc, root=root, tunning=True, fold=1, seed=seed)

            train_data = [train.__getitem__(i)[0] for i in range(len(train))]
            test_data = [test.__getitem__(i)[0] for i in range(len(test))]
            
            for trd_i, trd in enumerate(train_data):
                    for ted_i, ted in enumerate(test_data):
                            if torch.equal(trd, ted):
                                    exit()

def test_trainset():
    base = to.data.base_datasets.MNIST_base(root=root, download=True, transform=t_mnist)
    config = to.data.configure_division(base, n_openness, repeats, seed)

    for config_i, (kkc, uuc) in enumerate(config[0]):
        train, test = to.data.get_train_test(base, kkc, uuc, root=root, tunning=True, fold=1, seed=seed)

        train_labels = [train.__getitem__(i)[1] for i in range(len(train))]
        
        for trd_i, trd in enumerate(train_labels):
                #Czy na pewno w zbiorze treningowym nie znajdzie się UUC
                if (trd[-1]!=0):
                        exit()

def test_SVHN_config():
    base = to.data.base_datasets.SVHN_base(root=root, download=True, transform=t_svhn)
    config = to.data.configure_division(base, n_openness, repeats, seed)

    for config_i, (kkc, uuc) in enumerate(config[0]):
            train, test = to.data.get_train_test(base, kkc, uuc, root=root, tunning=True, fold=3, seed=seed)

            train_data = [train.__getitem__(i)[0] for i in range(len(train))]
            test_data = [test.__getitem__(i)[0] for i in range(len(test))]
            
            for trd_i, trd in enumerate(train_data):
                    for ted_i, ted in enumerate(test_data):
                            if torch.equal(trd, ted):
                                    exit()

def test_CIFAR10_config():
    base = to.data.base_datasets.CIFAR10_base(root=root, download=True, transform=t_cifar)
    config = to.data.configure_division(base, n_openness, repeats, seed)

    for config_i, (kkc, uuc) in enumerate(config[0]):
            train, test = to.data.get_train_test(base, kkc, uuc, root=root, tunning=True, fold=3, seed=seed)

            train_data = [train.__getitem__(i)[0] for i in range(len(train))]
            test_data = [test.__getitem__(i)[0] for i in range(len(test))]
            
            for trd_i, trd in enumerate(train_data):
                    for ted_i, ted in enumerate(test_data):
                            if torch.equal(trd, ted):
                                    exit()

def test_CIFAR100_config():
    base = to.data.base_datasets.CIFAR100_base(root=root, download=True, transform=t_cifar)
    config = to.data.configure_division(base, n_openness, repeats, seed)

    for config_i, (kkc, uuc) in enumerate(config[0]):
            train, test = to.data.get_train_test(base, kkc, uuc, root=root, tunning=True, fold=3, seed=seed)

            train_data = [train.__getitem__(i)[0] for i in range(len(train))]
            test_data = [test.__getitem__(i)[0] for i in range(len(test))]
            
            for trd_i, trd in enumerate(train_data):
                    for ted_i, ted in enumerate(test_data):
                            if torch.equal(trd, ted):
                                    exit()

def test_config_replicability():
    r_seed = 922
    r_openness = 10
    r_repeats = 10
   
    base = to.data.base_datasets.MNIST_base(root=root, download=True, transform=t_mnist)
    
    config1 = to.data.configure_division(base, r_openness, r_repeats, r_seed)
    config2 = to.data.configure_division(base, r_openness, r_repeats, r_seed)
    
    for i in range(len(config1)):
            #kkc
            assert(torch.equal(config1[0][i][0], config2[0][i][0]))
            #uuc
            assert(torch.equal(config1[0][i][1], config2[0][i][1]))
            #openness
            assert(torch.equal(config1[1][i], config2[1][i]))
            

def test_train_test_replicability():
    r_seed = 922
    r_kkc = [0,1,2]
    r_uuc = [3,4,5,6,7]
   
    base = to.data.base_datasets.MNIST_base(root=root, download=True, transform=t_mnist)
    
    train1, test1 = to.data.get_train_test(base, r_kkc, r_uuc, root=root, tunning=True, fold=1, seed=r_seed)
    train2, test2 = to.data.get_train_test(base, r_kkc, r_uuc, root=root, tunning=True, fold=1, seed=r_seed)

    for i in range(len(train1)):
            d1, t1 = train1.__getitem__(i)
            d2, t2 = train2.__getitem__(i)
            
            assert(torch.equal(d1, d2))
            assert(torch.equal(t1, t2))
            
    for i in range(len(test1)):
            d1, t1 = test1.__getitem__(i)
            d2, t2 = test2.__getitem__(i)
            
            assert(torch.equal(d1, d2))
            assert(torch.equal(t1, t2))

def test_tune_validation():
        base = to.data.base_datasets.MNIST_base(root=root, download=True, transform=t_mnist)
        kkc = [1,2]
        uuc = [0,3]

        for fold in range(n_splits):

                train_tunning, test_tunning = to.data.get_train_test(base, kkc, uuc, root=root, tunning=True, fold=fold, seed=seed)
                train_validation, test_validation = to.data.get_train_test(base, kkc, uuc, root=root, tunning=False, fold=fold, seed=seed)

                train_tunning_data = [train_tunning.__getitem__(i)[0] for i in range(len(train_tunning))]
                train_validation_data = [train_validation.__getitem__(i)[0] for i in range(len(train_validation))]
                                
                test_tunning_data = [test_tunning.__getitem__(i)[0] for i in range(len(test_tunning))]
                test_validation_data = [test_validation.__getitem__(i)[0] for i in range(len(test_validation))]

                validation_data = train_validation_data + test_validation_data
                tunning_data = train_tunning_data + test_tunning_data
                
                for vd_i, vd in enumerate(validation_data):
                        for td_i, td in enumerate(tunning_data):
                                if torch.equal(vd, td):
                                        exit()


from torch import nn

def flat_empty_lower_stack():
    """
    Returns empty flat architecture without any processing. 

    :rtype: torch.nn.Sequential
    :returns: Lower stack sequential architecture
    """
    
    return nn.Sequential()

def flat_fc_lower_stack(n_in_channels, n_out_channels, n_layers=1):
    """
    Returns simple flat architecture with single Fully Connected layer and Relu activatation. 

    :type n_in_channels: int
    :param n_in_channels: Input size
    
    :type n_out_channels: int
    :param n_out_channels: Output size

    :rtype: torch.nn.Sequential
    :returns: Lower stack sequential architecture
    """
    
    return nn.Sequential(
        nn.Linear(n_in_channels, n_out_channels),
        nn.ReLU(),
        *([nn.Linear(n_out_channels, n_out_channels),
        nn.ReLU()]*(n_layers-1)),
    )
    
def fc_lower_stack(depth, img_size_x, n_out_channels):
    """
    Returns simple architecture with single Fully Connected layer and Relu activatation. 

    :type depth: int
    :param depth: Number of color channels
    
    :type img_size_x: int
    :param img_size_x: Size of image in single axis
    
    :type n_out_channels: int
    :param n_out_channels: Size output

    :rtype: torch.nn.Sequential
    :returns: Lower stack sequential architecture
    """
    
    return nn.Sequential(
        nn.Flatten(),
        nn.Linear(depth*img_size_x*img_size_x, n_out_channels),
        nn.ReLU(),
    )
    
def osrci_lower_stack(depth, img_size_x, n_out_channels):
    """
    Returns architecture based on research in OSRCI method.

    :type depth: int
    :param depth: Number of color channels
    
    :type img_size_x: int
    :param img_size_x: Size of image in single axis
    
    :type n_out_channels: int
    :param n_out_channels: Size output

    :rtype: torch.nn.Sequential
    :returns: Lower stack sequential architecture
    """
    
    return nn.Sequential(
    nn.Conv2d(depth, 10, kernel_size=5),
    nn.MaxPool2d(2),
    nn.ReLU(),
    nn.Conv2d(10, 20, kernel_size=5),
    nn.Dropout2d(),
    nn.MaxPool2d(2),
    nn.ReLU(),
    nn.Flatten(),
    nn.Linear(_osrci_out_size(img_size_x), n_out_channels),
    nn.ReLU()
    )
    
def _osrci_out_size(x):
    x = x-4
    x = int(x/2)
    x = x-4
    x = int(x/2)
    x = x*x
    return 20*x

    
def alexNet32_lower_stack(n_out_channels): 
    """
    Returns modified architecture based on AlexNet, suitable for size (32 x 32 x 3) images.

    :type n_out_channels: int
    :param n_out_channels: Size output

    :rtype: torch.nn.Sequential
    :returns: Lower stack sequential architecture
    """
    return nn.Sequential(
            nn.Conv2d(3, 96, kernel_size=3), #30 x 30 x 96
            nn.ReLU(),
            nn.MaxPool2d(kernel_size=2, stride=2), # 15 x 15 x 96
            
            nn.Conv2d(96, 256, kernel_size=3), # 13 x 13 x 256
            nn.ReLU(),
                        
            nn.Conv2d(256, 384, kernel_size=3, padding=1), # 13 x 13 x 384
            nn.ReLU(),
            
            nn.Conv2d(384, 384, kernel_size=3, padding=1), # 13 x 13 x 384
            nn.ReLU(),
            
            nn.Conv2d(384, 256, kernel_size=3, padding=1), # 13 x 13 x 256
            nn.ReLU(),
            nn.MaxPool2d(kernel_size=3, stride=2), # 6 x 6 x 256
            
            nn.Dropout(),

            nn.Flatten(),
            
            nn.Linear(256 * 6 * 6, 4096),
            nn.ReLU(),
            
            nn.Dropout(),
            
            nn.Linear(4096, n_out_channels),
            nn.ReLU(),
            )


from .MNIST_base import MNIST_base
from .Omniglot_base import Omniglot_base
from .CIFAR10_base import CIFAR10_base
from .CIFAR100_base import CIFAR100_base
from .SVHN_base import SVHN_base
from .Madelon_base import Madelon_base
from .Moonway_base import Moonway_base

__all__ = [
    "MNIST_base",
    "Omniglot_base",
    "CIFAR10_base",
    "CIFAR100_base",
    "SVHN_base",
    "Madelon_base",
    "Moonway_base"
]



import torch
from torch import nn
from torchmetrics import Accuracy, ConfusionMatrix
from .base import OSRModule

class GSL(OSRModule):
    def __init__(self, lower_stack=None, n_known=3, sigma=3, n_generated=1.0, verbose=False, normal=False, threshold=None):
        super(GSL, self).__init__(lower_stack=lower_stack,
                                     n_known=n_known)
        self.sigma = sigma
        self.n_generated = n_generated
        self.randy = None
        self.verbose = verbose
        self.normal = normal
        self.threshold = threshold
        
        self.upper_stack = nn.Sequential(
            nn.Linear(64,32),
            nn.ReLU(),
            nn.Linear(32, self.n_known + 1)
        )
        
        
    def forward(self, X):
        representation = self.lower_stack(X)

        # Input background tension if representation requires grad        
        if representation.requires_grad:
            # Calculate representation
            mean = representation.mean(0)
            std = representation.std(0)
            if self.normal:
                rand_representation = torch.randn((int(self.n_generated*representation.shape[0]), representation.shape[1])) * std * self.sigma + mean
            else:
                rand_representation = torch.rand((int(self.n_generated*representation.shape[0]), representation.shape[1])) * std * self.sigma + mean
            
            # Join representations
            cat_represenation = torch.cat((representation, rand_representation))
            
            # Establish logits on joined
            logits = self.upper_stack(cat_represenation)
            
            # Prepare Randy
            self.randy = torch.zeros((rand_representation.shape[0], logits.shape[1]))
            self.randy[:,-1] = 1
        else:
            logits = self.upper_stack(representation)

        return logits
        
        
    def train(self, dataloader, loss_fn, optimizer, rand_kickstart=False):
        size = len(dataloader.dataset)
        for batch, (X, y) in enumerate(dataloader):
            
            if rand_kickstart:
                e_X = torch.rand_like(X)
                e_y = 1-y
                
                X = torch.cat([X,e_X])
                y = torch.cat([y,e_y])
                
            # Compute prediction and loss
            pred = self(X)        
            ry = torch.cat((y, self.randy))        
            loss = loss_fn(pred, ry)
            
            # Backpropagation
            optimizer.zero_grad()
            loss.backward()
            optimizer.step()

            if self.verbose:
                if batch % 100 == 0:
                    loss, current = loss.item(), batch * len(X)
                    print(f"loss: {loss:>7f}  [{current:>5d}/{size:>5d}]")

                
    def test(self, dataloader, loss_fn, conf=False): 
        inner_preds = []
        inner_preds_harry_potter = []
        inner_preds_overall = []
        outer_preds = []
        
        inner_targets = []
        outer_targets = []
        overall_targets = []
        
        # Define metrics for open set detection
        try:
            inner_metric = Accuracy(task="multiclass", num_classes=self.n_known, average='macro')
            overall_metric = Accuracy(task="multiclass", num_classes=self.n_known+1, average='macro')
        except:
            inner_metric = None
            overall_metric = None
            
        outer_metric = Accuracy(task='multiclass', num_classes=2, average='macro')

        with torch.no_grad():
            for X, y in dataloader:
                # Get y flat and known mask
                y_flat = y.argmax(1)
                known_mask = ~(y_flat == self.n_known)
                
                # Calculate logits for full batch
                logits = self(X)

                # # Establish inner preds and inner y
                inner_pp = nn.Softmax(dim=1)(logits[:,:-1])
                inner_pred = inner_pp.argmax(1)[known_mask]
                inner_target = y_flat[known_mask]
                
                inner_pp_overall = nn.Softmax(dim=1)(logits)
                inner_pred_overall = inner_pp_overall.argmax(1)

                # Establish outer pred
                outer_pred = (logits.argmax(1) != self.n_known).int()

                if self.threshold is not None:
                    low_support_mask = inner_pp_overall.max(1)[0] < self.threshold
                    
                    inner_pred_overall[low_support_mask] = self.n_known                    
                    outer_pred[low_support_mask] = 0
                
                inner_pred_harry_potter = inner_pred_overall[known_mask]

                outer_target = (y_flat != self.n_known).int()
 
                # Store predictions
                inner_preds.append(inner_pred)
                inner_preds_harry_potter.append(inner_pred_harry_potter)
                inner_preds_overall.append(inner_pred_overall)
                inner_targets.append(inner_target)
                overall_targets.append(y_flat)
                
                outer_preds.append(outer_pred)
                outer_targets.append(outer_target)
                
        inner_targets = torch.cat(inner_targets)
        overall_targets = torch.cat(overall_targets)
        inner_preds = torch.cat(inner_preds)
        inner_preds_harry_potter = torch.cat(inner_preds_harry_potter)
        inner_preds_overall = torch.cat(inner_preds_overall)
        
        outer_targets = torch.cat(outer_targets)
        outer_preds = torch.cat(outer_preds)
        
        # Calculate scores
        if inner_metric is not None:
            inner_score = inner_metric(inner_preds, inner_targets)
            inner_score_harry = overall_metric(inner_preds_harry_potter, inner_targets)
            inner_score_overall = overall_metric(inner_preds_overall, overall_targets)
        else:
            inner_score = torch.nan
            inner_score_harry = torch.nan
            inner_score_overall = torch.nan
            
        outer_score = outer_metric(outer_preds, outer_targets)

        if conf==True:
            c_inner = ConfusionMatrix(task="multiclass", num_classes=self.n_known)(inner_preds, inner_targets)
            c_outer = ConfusionMatrix(task="binary")(outer_preds, outer_targets)
            c_hp = ConfusionMatrix(task="multiclass", num_classes=self.n_known+1)(inner_preds_harry_potter, inner_targets)
            c_overall = ConfusionMatrix(task="multiclass", num_classes=self.n_known+1)(inner_preds_overall, overall_targets)
            return inner_score, outer_score, inner_score_harry, inner_score_overall, c_inner, c_outer, c_hp, c_overall
           
        if self.verbose:
            print('Inner metric : %.3f' % inner_score)
            print('Outer metric : %.3f' % outer_score)
            
        return inner_score, outer_score, inner_score_harry, inner_score_overall


    def predict(self, X): 
        with torch.no_grad():

            logits = self(X)
            pred_overall = nn.Softmax(dim=1)(logits).argmax(1)

            return pred_overall
            


import os.path
import pickle
from typing import Any, Callable, Optional, Tuple
import numpy as np
from PIL import Image

from torchvision.datasets import VisionDataset
from torchvision.datasets.utils import check_integrity, download_and_extract_archive


class CIFAR10_base(VisionDataset):
    """
    `CIFAR10 <https://www.cs.toronto.edu/~kriz/cifar.html>`_ Dataset.
    
    Base implementation from torch library. No split to train/test data.
    """
    
    base_folder = "cifar-10-batches-py"
    url = "https://www.cs.toronto.edu/~kriz/cifar-10-python.tar.gz"
    filename = "cifar-10-python.tar.gz"
    tgz_md5 = "c58f30108f718f92721af3b95e74349a"
    train_list = [
        ["data_batch_1", "c99cafc152244af753f735de768cd75f"],
        ["data_batch_2", "d4bba439e000b95fd0a9bffe97cbabec"],
        ["data_batch_3", "54ebc095f3ab1f0389bbae665268c751"],
        ["data_batch_4", "634d18415352ddfa80567beed471001a"],
        ["data_batch_5", "482c414d41f54cd18b22e5b47cb7c3cb"],
    ]

    test_list = [
        ["test_batch", "40351d587109b95175f43aff81a1287e"],
    ]
    meta = {
        "filename": "batches.meta",
        "key": "label_names",
        "md5": "5ff9c542aee3614f3951f8cda6e48888",
    }

    def __init__(
        self,
        root: str,
        transform: Optional[Callable] = None,
        target_transform: Optional[Callable] = None,
        download: bool = False,
    ) -> None:

        super().__init__(root, transform=transform, target_transform=target_transform)

        if download:
            self.download()

        if not self._check_integrity():
            raise RuntimeError("Dataset not found or corrupted. You can use download=True to download it")

        data_train = []
        data_test = []
        
        targets_train = []
        targets_test = []

        #Train data
        for file_name, checksum in self.train_list:
            file_path = os.path.join(self.root, self.base_folder, file_name)
            with open(file_path, "rb") as f:
                entry = pickle.load(f, encoding="latin1")
                data_train.append(entry["data"])
                if "labels" in entry:
                    targets_train.extend(entry["labels"])
                else:
                    targets_train.extend(entry["fine_labels"])

        data_train = np.vstack(data_train).reshape(-1, 3, 32, 32)
        data_train = data_train.transpose((0, 2, 3, 1))  # convert to HWC

        #Test data
        for file_name, checksum in self.test_list:
            file_path = os.path.join(self.root, self.base_folder, file_name)
            with open(file_path, "rb") as f:
                entry = pickle.load(f, encoding="latin1")
                data_test.append(entry["data"])
                if "labels" in entry:
                    targets_test.extend(entry["labels"])
                else:
                    targets_test.extend(entry["fine_labels"])

        data_test= np.vstack(data_test).reshape(-1, 3, 32, 32)
        data_test = data_test.transpose((0, 2, 3, 1))  # convert to HWC

        self.data = np.concatenate((data_train, data_test))
        self.targets = np.concatenate((targets_train, targets_test))
        
        
    def __getitem__(self, index: int) -> Tuple[Any, Any]:
        img, target = self.data[index], self.targets[index]

        # doing this so that it is consistent with all other datasets
        # to return a PIL Image
        img = Image.fromarray(img)

        if self.transform is not None:
            img = self.transform(img)

        if self.target_transform is not None:
            target = self.target_transform(target)

        return img, target

    def __len__(self) -> int:
        return len(self.data)
    
    def _n_classes(self) -> int:
        return len(np.unique(self.targets))

    def _check_integrity(self) -> bool:
        for filename, md5 in self.train_list + self.test_list:
            fpath = os.path.join(self.root, self.base_folder, filename)
            if not check_integrity(fpath, md5):
                return False
        return True

    def download(self) -> None:
        if self._check_integrity():
            print("Files already downloaded and verified")
            return
        download_and_extract_archive(self.url, self.root, filename=self.filename, md5=self.tgz_md5)

    def extra_repr(self) -> str:
        split = "Train" if self.train is True else "Test"
        return f"Split: {split}"



import torch
from torch import nn
from torchmetrics import ConfusionMatrix, Accuracy
import itertools
from .base import OSRModule
import numpy as np
from numpy.matlib import repmat

class OverlaySoftmax(OSRModule):
    """
    Implementation of OverlaySoftmax method

    :type lower_stack: nn.Sequential
    :param lower_stack: Network architecture of lower_stack
    
    :type n_known: int
    :param n_known: Number of known classes
    
    """
    def __init__(self, 
                 lower_stack=None, 
                 n_known=3, 
                 threshold=None, 
                 size=0.5,
                 shuffle=False,
                 n_mixed=2):
        super(OverlaySoftmax, self).__init__(lower_stack=lower_stack,
                                      n_known=n_known)
        self.randy = None
        self.size = size
        self.threshold = threshold
        self.shuffle = shuffle
        self.n_mixed = n_mixed
        
        self.upper_stack = nn.Sequential(
            nn.Linear(64, self.n_known+1)
        )
        
    def train(self, dataloader, loss_fn, optimizer):
        for batch, (X, y) in enumerate(dataloader):
            # Establish number of target samples            
            n_target_samples = int(np.ceil(X.shape[0] * self.size))
            
            # Calculate known class distribution and list present classes
            distribution = y.sum(0).detach().numpy().astype(int)
            present = np.where(distribution != 0)[0]
            
            # Establish all combinations and enhance it to the target number of samples
            pairing = np.array(list(itertools.combinations(present, self.n_mixed)))         
            n_review = len(pairing)
            
            # Secure on single-class batches
            if n_review > 0:
                n_cycles = np.ceil(n_target_samples / n_review).astype(int)
                pairing = repmat(pairing, n_cycles, 1)[:n_target_samples]
                
                # Shuffle pairings if necessary
                if self.shuffle:
                    np.random.shuffle(pairing)
                
                # Prepare storage
                overlayed_X = []
                
                # Iterate pairing
                for pid, elements in enumerate(pairing):
                    # Select sample sublocation
                    locs = pid % distribution[elements]
                    
                    # Establish sample location
                    ids = [np.where(y[:,element] == 1)[0][loc] for (element, loc) in zip(elements, locs)]
                    
                    # Integrate and store sample
                    x = X[ids].mean(0)                
                    overlayed_X.append(x)
                    
                overlayed_X = torch.stack(overlayed_X)            
                overlayed_y = torch.zeros((overlayed_X.shape[0], self.n_known+1))
                overlayed_y[:,-1] = 1
                
                _X = torch.cat((X, overlayed_X))
                _y = torch.cat((y, overlayed_y))
            else:
                _X = X
                _y = y
            
            # Compute prediction and loss
            pred = self(_X)
            loss = loss_fn(pred, _y)
            
            # Backpropagation
            optimizer.zero_grad()
            loss.backward()
            optimizer.step()
                
    def test(self, dataloader, loss_fn, conf=False):
        inner_preds = []
        inner_preds_hp = []
        inner_preds_overall = []
        outer_preds = []
        
        inner_targets = []
        outer_targets = []
        overall_targets = []
        
        # Define metric for open set detection
        try:
            inner_metric = Accuracy(task="multiclass", num_classes=self.n_known, average='macro')
            overall_metric = Accuracy(task="multiclass", num_classes=self.n_known+1, average='macro')
        except:
            inner_metric = None
            
        outer_metric = Accuracy(task='multiclass', num_classes=2, average='macro')

        with torch.no_grad():
            for X, y in dataloader:
                # Get y flat and known mask
                y_flat = y.argmax(1)
                known_mask = ~(y_flat == self.n_known)
                
                # Calculate logits for full batch
                logits = self(X)

                # # Establish inner preds and inner y
                inner_pp = nn.Softmax(dim=1)(logits[:,:-1])
                inner_pred = inner_pp.argmax(1)[known_mask]
                inner_target = y_flat[known_mask]
                
                inner_pp_overall = nn.Softmax(dim=1)(logits)
                overall_pred = inner_pp_overall.argmax(1)

                # Establish outer pred
                outer_pred = (logits.argmax(1) != self.n_known).int()

                if self.threshold is not None:
                    low_support_mask = inner_pp_overall.max(1)[0] < self.threshold
                    
                    overall_pred[low_support_mask] = self.n_known                    
                    outer_pred[low_support_mask] = 0
                
                inner_pred_hp = overall_pred[known_mask]

                outer_target = (y_flat != self.n_known).int()
                # print(torch.unique(outer_pred, return_counts=True))
                
                # Store predictions
                inner_preds.append(inner_pred)
                inner_preds_hp.append(inner_pred_hp)
                inner_preds_overall.append(overall_pred)

                inner_targets.append(inner_target)
                overall_targets.append(y_flat)
                
                outer_preds.append(outer_pred)
                outer_targets.append(outer_target)
                
        inner_targets = torch.cat(inner_targets)
        overall_targets = torch.cat(overall_targets)
        inner_preds = torch.cat(inner_preds)
        inner_preds_hp = torch.cat(inner_preds_hp)
        inner_preds_overall = torch.cat(inner_preds_overall)
        
        outer_targets = torch.cat(outer_targets)
        outer_preds = torch.cat(outer_preds)
        
        # Calculate scores
        if inner_metric is not None:
            inner_score = inner_metric(inner_preds, inner_targets)
            inner_score_hp = overall_metric(inner_preds_hp, inner_targets)
            inner_score_overall = overall_metric(inner_preds_overall, overall_targets)
        else:
            inner_score = torch.nan
            inner_score_hp = torch.nan
            inner_score_overall = torch.nan
            
        outer_score = outer_metric(outer_preds, outer_targets)
        
        if conf==True:
            c_inner = ConfusionMatrix(task="multiclass", num_classes=self.n_known)(inner_preds, inner_targets)
            c_outer = ConfusionMatrix(task="binary")(outer_preds, outer_targets)
            c_hp = ConfusionMatrix(task="multiclass", num_classes=self.n_known+1)(inner_preds_hp, inner_targets)
            c_overall = ConfusionMatrix(task="multiclass", num_classes=self.n_known+1)(inner_preds_overall, overall_targets)
            return inner_score, outer_score, inner_score_hp, inner_score_overall, c_inner, c_outer, c_hp, c_overall
            
        if self.verbose:   
            print('Inner metric : %.3f' % inner_score)
            print('Outer metric : %.3f' % outer_score)
            
        return inner_score, outer_score, inner_score_hp, inner_score_overall


    def predict(self, X):
        with torch.no_grad():
            logits = self(X)
            
            overall_pred = nn.Softmax(dim=1)(logits).argmax(1)
            outer_pred = (nn.Softmax(dim=1)(logits).max(1).values > self.epsilon).int()
            overall_pred[outer_pred==0]=self.n_known
            
            return overall_pred
                
        


import numpy as np
from sklearn.datasets import make_moons
from torch.utils.data import Dataset
from typing import Any, Tuple
from sklearn.utils import shuffle

def make_moonway(n_circles = 5, sigma = .1, n_samples_moons = 100, n_samples_per_circle = 100, mode = 'linear',
                 random_state=None, padding=2):
    np.random.seed(random_state)
    
    n_samples = n_samples_moons + n_circles * n_samples_per_circle
    noise = np.random.normal(0, sigma, size=(n_samples, 2))

    m_X, m_y = make_moons(n_samples=n_samples_moons)
    m_X -= [.5,0]

    if mode == 'linear':
        circle_radiuses = padding + np.arange(n_circles)
    elif mode == 'log':    
        circle_radiuses = np.logspace(padding, n_circles, n_circles)
    else:
        raise Exception("Only linear and log modes available")
    
    angles = np.random.uniform(0, np.pi*2, (n_samples_per_circle, n_circles))

    c_X = np.array([np.sin(angles) * circle_radiuses[None,:],
                    np.cos(angles) * circle_radiuses[None,:]])
    c_y = np.ones((n_samples_per_circle, n_circles)) * (np.arange(n_circles)+2)[None, :]

    c_X = c_X.reshape(2,-1).T
    c_y = c_y.reshape(-1)

    X = np.concatenate((c_X, m_X)) + noise
    y = np.concatenate((c_y, m_y))
    
    X, y = shuffle(X, y, random_state=random_state)
    
    return X, y

class Moonway_base(Dataset):
    def __init__(self, **kwargs):
        self.kwargs = kwargs
        self.data, self.targets = self._load_data()

    def _load_data(self):
        data, targets = make_moonway(**self.kwargs)
        return data.astype(np.float32), targets.astype(np.int64)

    def __getitem__(self, index: int) -> Tuple[Any, Any]:
        return self.data[index], self.targets[index]

    def __len__(self) -> int:
        return len(self.targets)
    
    def _n_classes(self) -> int:
        return len(np.unique(self.targets))



from .CIFAR10_base import CIFAR10_base

class CIFAR100_base(CIFAR10_base):
    """
    `CIFAR100 <https://www.cs.toronto.edu/~kriz/cifar.html>`_ Dataset.
    
    Base implementation from torch library. No split to train/test data.
    """

    base_folder = "cifar-100-python"
    url = "https://www.cs.toronto.edu/~kriz/cifar-100-python.tar.gz"
    filename = "cifar-100-python.tar.gz"
    tgz_md5 = "eb9058c3a382ffc7106e4002c42a8d85"
    train_list = [
        ["train", "16019d7e3df5f24257cddd939b257f8d"],
    ]

    test_list = [
        ["test", "f0ef6b0ae62326f3e7ffdfab6717acfc"],
    ]
    meta = {
        "filename": "meta",
        "key": "fine_label_names",
        "md5": "7973b15100ade9c7d40fb424638fde48",
    }


import os.path
from typing import Any, Callable, Optional, Tuple
import numpy as np
from PIL import Image
import scipy.io as sio

from torchvision.datasets import VisionDataset
from torchvision.datasets.utils import check_integrity, download_url



class SVHN_base(VisionDataset):
    """
    `SVHN <http://ufldl.stanford.edu/housenumbers/>`_ Dataset.
        
    Base implementation from torch library. No split to train/test data.
    """

    split_list = {
        "train": [
            "http://ufldl.stanford.edu/housenumbers/train_32x32.mat",
            "train_32x32.mat",
            "e26dedcc434d2e4c54c9b2d4a06d8373",
        ],
        "test": [
            "http://ufldl.stanford.edu/housenumbers/test_32x32.mat",
            "test_32x32.mat",
            "eb5a983be6a315427106f1b164d9cef3",
        ],
        # "extra": [
        #     "http://ufldl.stanford.edu/housenumbers/extra_32x32.mat",
        #     "extra_32x32.mat",
        #     "a93ce644f1a588dc4d68dda5feec44a7",
        # ],
    }

    def __init__(
        self,
        root: str,
        transform: Optional[Callable] = None,
        target_transform: Optional[Callable] = None,
        download: bool = False,
    ) -> None:
        super().__init__(root, transform=transform, target_transform=target_transform)

        if download:
            self.download()

        if not self._check_integrity():
            raise RuntimeError("Dataset not found or corrupted. You can use download=True to download it")

        # reading(loading) mat file as array
        loaded_mat = [sio.loadmat(os.path.join(self.root, self.split_list[split][1])) for split in self.split_list.keys()]
        
        self.data = []
        self.labels = []
        
        for lm in loaded_mat:
            data = lm["X"]
            # loading from the .mat file gives an np array of type np.uint8
            # converting to np.int64, so that we have a LongTensor after
            # the conversion from the numpy array
            # the squeeze is needed to obtain a 1D tensor
            labels = lm["y"].astype(np.int64).squeeze()

            # the svhn dataset assigns the class label "10" to the digit 0
            # this makes it inconsistent with several loss functions
            # which expect the class labels to be in the range [0, C-1]
            np.place(labels, labels == 10, 0)
            data = np.transpose(data, (3, 2, 0, 1))
            
            self.data.append(data)
            self.labels.append(labels)
        
        self.data = np.concatenate(self.data)
        self.labels = np.concatenate(self.labels)
        

    def __getitem__(self, index: int) -> Tuple[Any, Any]:
        img, target = self.data[index], int(self.labels[index])

        # doing this so that it is consistent with all other datasets
        # to return a PIL Image
        img = Image.fromarray(np.transpose(img, (1, 2, 0)))

        if self.transform is not None:
            img = self.transform(img)

        if self.target_transform is not None:
            target = self.target_transform(target)

        return img, target

    def __len__(self) -> int:
        return len(self.data)
    
    def _n_classes(self) -> int:
        return len(np.unique(self.labels))

    def _check_integrity(self) -> bool:
        root = self.root
        for split in self.split_list.keys():
            md5 = self.split_list[split][2]
            fpath = os.path.join(root, self.split_list[split][1])
            return check_integrity(fpath, md5)

    def download(self) -> None:
        for split in self.split_list.keys():
            md5 = self.split_list[split][2]
            download_url(self.split_list[split][0], self.root, self.split_list[split][1], md5)

    def extra_repr(self) -> str:
        return "Split: {split}".format(**self.__dict__)


import torch
from ..data.DataWrapper import DataWrapper
from ..data.OutlierDataset import OutlierDataset
import numpy as np
import copy

def configure_division_outlier(base_dataset, outlier_dataset, repeats, n_openness=None, seed=None, min_known_classes=2):
        """
        Method for obtaining configurations for OSR model evaluation using Outlier protocol. KKC come from base_dataset, UUC from outlier_dataset.

        :type base_dataset: VisionDataset
        :param base_dataset: Dataset describing KKC instances
        
        :type outlier_dataset: VisionDataset
        :param outlier_dataset: Dataset describing UUC instances
        
        :type repeats: int
        :param repeats: Number of randol selections of classes for single openness (KKC/UUC class cardinality)
        
        :type n_openness: int
        :param n_openness: Number of KKC/UUC class cardinality to generate
        
        :type seed: int
        :param seed: Random state
        
        :type min_known_classes: int
        :param min_known_classes: Minimum number of known classes
        
        :rtype: List
        :returns: Lit of dataset configurations -- each containing sets of KKC and UUC -- and their Openness
        """
        trng = torch.random.manual_seed(seed)

        # Get number of classes from base and outlier datasets
        n_classes_base = (base_dataset._n_classes())
        n_classes_out = (outlier_dataset._n_classes())

        # Tables for quantities of KKC and UUC and Opennes dependent of quantities
        openness_n_classes = []
        openness = []
        
        for i in range(min_known_classes,n_classes_base):
                for j in range(1,n_classes_out):
                      openness_n_classes.append([i,j])  
                      openness.append(1 - np.sqrt((2*i)/((i*2)+j)))

        openness = torch.tensor(openness)
        openness_n_classes = torch.tensor(openness_n_classes)
        
        if n_openness is not None:
                # Randomly select n_openness configurations
                # Higher selection probability to configurations with larger number of classes
                p = openness_n_classes.sum(1)
                p = p / p.sum()
                rand_indexes = p.multinomial(num_samples=n_openness, replacement=False, generator=trng)
                
                # Select chosen openness and configurations
                op_choice = openness[rand_indexes]
                op_n_classes_choice = openness_n_classes[rand_indexes]
        else:
                # In n_openness is None select all configurations
                op_choice = openness
                op_n_classes_choice = openness_n_classes

        # Now randomly assign classes
        repeats_config =[]
        for kkc_n, uuc_n in op_n_classes_choice:    
                for r in range(repeats):
                        all_kkc = torch.arange(n_classes_base, dtype=float)
                        kkc = all_kkc.multinomial(num_samples=kkc_n, replacement=False, generator=trng)
                        
                        all_uuc = torch.arange(n_classes_out, dtype=float)
                        uuc = all_uuc.multinomial(num_samples=uuc_n, replacement=False, generator=trng)
                        
                        repeats_config.append((kkc, uuc))
        
        return repeats_config, op_choice


def get_train_test_outlier(base_dataset, outlier_dataset, kkc_indexes, uuc_indexes, root, tunning, fold, seed=1410, n_folds=5):
        """
        Method for obtaining Cross-validation folds using Outlier protocol (KKC obtained from base_dataset and UUC from outlier_dataset).

        :type base_dataset: VisionDataset
        :param base_dataset: Dataset describing KKC instances
        
        :type outlier_dataset: VisionDataset
        :param outlier_dataset: Dataset describing UUC instances
        
        :type kkc_indexes: List
        :param kkc_indexes: List of labels constituting Known Classes        
        
        :type uuc_indexes: List
        :param uuc_indexes: List of labels constituting Unknown Classes
        
        :type root: string
        :param root: Datasets folder
        
        :type tunning: boolean
        :param tunning: Flag. If True will split 10% of data for tunning, otherwise will split 90% of data.
                
        :type fold: int
        :param fold: Fold index
        
        :type n_folds: int
        :param n_folds: Number of folds
        
        :type seed: int
        :param seed: Random state
        
        :rtype: List
        :returns: Train dataset, Test dataset
        """
        
        trng = torch.random.manual_seed(seed)

        # Create dataset containing only KKC
        data = DataWrapper(root=root,
                       base_dataset=base_dataset,
                       indexes='all',
                       get_classes=kkc_indexes,
                       known_classes=kkc_indexes,
                       return_only_known=True,
                       onehot=True,
                       onehot_num_classes=len(kkc_indexes)+1
                       )
        
        labels = np.unique(data.original_targets)
        
        # Portion of data for tunning
        proportion = .1
        
        # First tuning, later validation
        division_probability = torch.ones(n_folds*2) * (1-proportion) / n_folds
        division_probability[:n_folds] = proportion / n_folds
        
        division_mask = np.zeros_like(data.targets)
        
        for label in labels:
                class_mask = data.original_targets == label

                division_mask[class_mask] = division_probability.multinomial(class_mask.sum(0), replacement=True, generator=trng)                
                
        if tunning:
                test_indexes = division_mask == fold
                train_indexes = division_mask < n_folds
                
                train_indexes = train_indexes * ~test_indexes
        else:
                test_indexes = division_mask == (fold+n_folds)
                train_indexes = division_mask >= n_folds
                
                train_indexes = train_indexes * ~test_indexes

        data_test = copy.deepcopy(data)
        data_train = copy.deepcopy(data)
                
        data_test.reindex(np.where(test_indexes)[0])
        data_train.reindex(np.where(train_indexes)[0])
        
        outliers = DataWrapper(root=root,
                       base_dataset=outlier_dataset,
                       indexes='all',
                       get_classes=uuc_indexes,
                       known_classes=uuc_indexes,
                       return_only_known=True,
                       onehot=False,
                       onehot_num_classes=None
                       )
        
        # Assign label to UUC objects
        outliers.targets = [torch.zeros(len(kkc_indexes)+1, dtype=torch.float)
                                         .scatter_(0, torch.tensor(len(kkc_indexes)), value=1)
                                         for t in outliers.targets]
        
        data_test = OutlierDataset(root, data_test, outliers, shuffle=False, onehot=False, unknown_label=len(kkc_indexes))
        
        return data_train, data_test



import codecs
import os
import os.path
import sys
from typing import Any, Callable, Optional, Tuple, Dict
from urllib.error import URLError

import numpy as np
import torch
from PIL import Image

from torchvision.datasets import VisionDataset
from torchvision.datasets.utils import check_integrity, download_and_extract_archive


class MNIST_base(VisionDataset):
    """
    `MNIST <http://yann.lecun.com/exdb/mnist/>`_ Dataset.
    
    Base implementation from torch library. No split to train/test data.
    """
    
    mirrors = [
        "http://yann.lecun.com/exdb/mnist/",
        "https://ossci-datasets.s3.amazonaws.com/mnist/",
    ]

    resources = [
        ("train-images-idx3-ubyte.gz", "f68b3c2dcbeaaa9fbdd348bbdeb94873"),
        ("train-labels-idx1-ubyte.gz", "d53e105ee54ea40749a09fcbcd1e9432"),
        ("t10k-images-idx3-ubyte.gz", "9fb629c4189551a2d022fa330f9573f3"),
        ("t10k-labels-idx1-ubyte.gz", "ec29112dd5afa0611ce80d1b7f02629c"),
    ]

    training_file = "training.pt"
    test_file = "test.pt"

    def __init__(
        self,
        root: str,
        transform: Optional[Callable] = None,
        target_transform: Optional[Callable] = None, #has to return int
        download: bool = False,
    ) -> None:
        super().__init__(root, transform=transform, target_transform=target_transform)

        if self._check_legacy_exist():
            self.data, self.targets = self._load_legacy_data()
            return

        if download:
            self.download()

        if not self._check_exists():
            raise RuntimeError("Dataset not found. You can use download=True to download it")

        self.data, self.targets = self._load_data()

    def _check_legacy_exist(self):
        processed_folder_exists = os.path.exists(self.processed_folder)
        if not processed_folder_exists:
            return False

        return all(
            check_integrity(os.path.join(self.processed_folder, file)) for file in (self.training_file, self.test_file)
        )

    def _load_legacy_data(self):
        # This is for BC only. We no longer cache the data in a custom binary, but simply read from the raw data
        # directly.
        data_file = self.training_file if self.train else self.test_file
        return torch.load(os.path.join(self.processed_folder, data_file))

    def _load_data(self):
        #Load data
        data_train = read_image_file(os.path.join(self.raw_folder, "train-images-idx3-ubyte"))
        data_test = read_image_file(os.path.join(self.raw_folder, "t10k-images-idx3-ubyte"))
        data = torch.cat((data_train,data_test))
        
        #Load labels
        targets_train = read_label_file(os.path.join(self.raw_folder, "train-labels-idx1-ubyte"))
        targets_test = read_label_file(os.path.join(self.raw_folder, "t10k-labels-idx1-ubyte"))
        targets = torch.cat((targets_train,targets_test))
        
        return data, targets

    def __getitem__(self, index: int) -> Tuple[Any, Any]:
        img, target = self.data[index], int(self.targets[index])

        # doing this so that it is consistent with all other datasets
        # to return a PIL Image
        img = Image.fromarray(img.numpy(), mode="L")

        if self.transform is not None:
            img = self.transform(img)

        if self.target_transform is not None:
            target = self.target_transform(target)

        return img, target

    def __len__(self) -> int:
        return len(self.targets)
    
    def _n_classes(self) -> int:
        return len(np.unique(self.targets))
    
    @property
    def raw_folder(self) -> str:
        return os.path.join(self.root, self.__class__.__name__, "raw")

    @property
    def processed_folder(self) -> str:
        return os.path.join(self.root, self.__class__.__name__, "processed")

    @property
    def class_to_idx(self) -> Dict[str, int]:
        return {_class: i for i, _class in enumerate(self.classes)}


    def _check_exists(self) -> bool:
        return all(
            check_integrity(os.path.join(self.raw_folder, os.path.splitext(os.path.basename(url))[0]))
            for url, _ in self.resources
        )

    def download(self) -> None:
        if self._check_exists():
            return

        os.makedirs(self.raw_folder, exist_ok=True)

        # download files
        for filename, md5 in self.resources:
            for mirror in self.mirrors:
                url = f"{mirror}{filename}"
                try:
                    print(f"Downloading {url}")
                    download_and_extract_archive(url, download_root=self.raw_folder, filename=filename, md5=md5)
                except URLError as error:
                    print(f"Failed to download (trying next):\n{error}")
                    continue
                finally:
                    print()
                break
            else:
                raise RuntimeError(f"Error downloading {filename}")

def get_int(b: bytes) -> int:
    return int(codecs.encode(b, "hex"), 16)


SN3_PASCALVINCENT_TYPEMAP = {
    8: torch.uint8,
    9: torch.int8,
    11: torch.int16,
    12: torch.int32,
    13: torch.float32,
    14: torch.float64,
}


def read_sn3_pascalvincent_tensor(path: str, strict: bool = True) -> torch.Tensor:
    #Read a SN3 file in "Pascal Vincent" format (Lush file 'libidx/idx-io.lsh').
    # Argument may be a filename, compressed filename, or file object.
    
    # read
    with open(path, "rb") as f:
        data = f.read()
    # parse
    magic = get_int(data[0:4])
    nd = magic % 256
    ty = magic // 256
    assert 1 <= nd <= 3
    assert 8 <= ty <= 14
    torch_type = SN3_PASCALVINCENT_TYPEMAP[ty]
    s = [get_int(data[4 * (i + 1) : 4 * (i + 2)]) for i in range(nd)]

    num_bytes_per_value = torch.iinfo(torch_type).bits // 8
    # The MNIST format uses the big endian byte order. If the system uses little endian byte order by default,
    # we need to reverse the bytes before we can read them with torch.frombuffer().
    needs_byte_reversal = sys.byteorder == "little" and num_bytes_per_value > 1
    parsed = torch.frombuffer(bytearray(data), dtype=torch_type, offset=(4 * (nd + 1)))
    if needs_byte_reversal:
        parsed = parsed.flip(0)

    assert parsed.shape[0] == np.prod(s) or not strict
    return parsed.view(*s)


def read_label_file(path: str) -> torch.Tensor:
    x = read_sn3_pascalvincent_tensor(path, strict=False)
    if x.dtype != torch.uint8:
        raise TypeError(f"x should be of dtype torch.uint8 instead of {x.dtype}")
    if x.ndimension() != 1:
        raise ValueError(f"x should have 1 dimension instead of {x.ndimension()}")
    return x.long()


def read_image_file(path: str) -> torch.Tensor:
    x = read_sn3_pascalvincent_tensor(path, strict=False)
    if x.dtype != torch.uint8:
        raise TypeError(f"x should be of dtype torch.uint8 instead of {x.dtype}")
    if x.ndimension() != 3:
        raise ValueError(f"x should have 3 dimension instead of {x.ndimension()}")
    return x


from os.path import join
from typing import Any, Tuple
import numpy as np
from ..utils.base import onehot_bg
from torchvision.datasets import VisionDataset


class OutlierDataset(VisionDataset):
    """
    OutlierDataset for combining instances from (known) training and (unknown) outlier data.
    
    :type root: string
    :param root: Data directory.
    
    :type dataset: VisionDataset
    :param dataset: Dataset of known-class testing examples.

    :type outliers: VisionDataset
    :param outliers: Dataset of unknown-class testing examples, which will be labeled as unknowns.
    
    :type shuffle: boolean
    :param shuffle: If True, the final data will be shuffled.
    
    :type random_state: int
    :param random_state: Random state (for shuffle).
    
    :type unknown_label: int
    :param unknown_label: Label with which the unknowns will be marked.

    :type onehot: boolean
    :param onehot: If True will perform one-hot encoding on labels.
    
    :type onehot_num_classes: int
    :param onehot_num_classes: Number of classes for one-hot encoding (in case outlier data is generated for testing). 
    """
    def __init__(
        self,
        root: str,
        dataset,
        outliers,
        shuffle: bool = True,
        random_state: int = None,
        unknown_label: int = None,
        onehot = False,
        onehot_num_classes = None
        ) -> None:
        super().__init__(join(root))
        
        self.dataset = dataset
        self.outliers = outliers
        self.shuffle = shuffle
        self.unknown_label = unknown_label
        self.onehot = onehot
        self.onehot_num_classes = onehot_num_classes
        
        if random_state is not None:
            np.random.seed(random_state)

        #Combine two datasets
        data = []
        targets = []
        for i in range(len(self.dataset)):
            d, t  = self.dataset.__getitem__(i)
            data.append(d)
            targets.append(t)
        for i in range(len(self.outliers)):
            d, t = self.outliers.__getitem__(i)
            data.append(d)
            targets.append(t)
            
        if self.shuffle:
            order = np.random.permutation(len(targets))
            data = np.array(data)[order]
            targets = np.array(targets)[order]
            
        self.data = data 
        self.targets = targets
    
    def __len__(self) -> int:
        return len(self.targets)

    def __getitem__(self, index: int) -> Tuple[Any, Any]:
        d, t = self.data[index], self.targets[index]
        if self.onehot:
            transform = onehot_bg(self.onehot_num_classes)
            t = transform(t)
        return d,t

    def _num_classes(self):
        return len(np.unique(self.targets))


from .base import (onehot_bg, inverse_transform, grayscale_transform, get_openmax_epsilon, get_softmax_epsilon)

__all__ = [
    "onehot_bg",
    "inverse_transform",
    "grayscale_transform",
    "get_openmax_epsilon",
    "get_softmax_epsilon"
]



from typing import Any, Tuple
import numpy as np
import torch
import copy
from ..utils.base import onehot_bg
from itertools import compress
from operator import itemgetter 

from torchvision.datasets import VisionDataset

class DataWrapper(VisionDataset):
    """
    DataWrapper for base datasets.
    
    :type root: string
    :param root: Data directory.
    
    :type base_dataset: VisionDataset
    :param base_dataset: base dataset implementing __getitem__ function.
    
    :type indexes: List
    :param indexes: Indexes of wrapped dataset objects that will be returned.
    
    :type get_classes: List
    :param get_classes: Considered class indexes from base dataset (known + unknown).
    
    :type known_classes: List
    :param known_classes: Considered known class indexes from base dataset.
    
    :type return_only_known: boolean
    :param return_only_known: If True will return only known instances (for training). If False will assign new class index (equal to the number of classes) to unknown samples, and return them with known data (for testing).

    :type onehot: boolean
    :param onehot: If True will perform one-hot encoding on labels.
    
    :type onehot_num_classes: int
    :param onehot_num_classes: Number of classes for one-hot encoding (in case outlier data is generated for testing). 
    """
    
    def __init__(
        self,
        root: str,
        base_dataset,
        get_classes,        # Known + Unknown
        known_classes,
        return_only_known,  # For training
        indexes = 'all',    # For k-fold
        onehot = False,
        onehot_num_classes = None
    ) -> None:
        super().__init__(root)
        
        self.base_dataset = base_dataset
        self.indexes = indexes
        self.get_classes = get_classes
        self.known_classes = known_classes
        self.return_only_known = return_only_known
        self.onehot = onehot
        self.onehot_num_classes = onehot_num_classes
        
        self.new_index_mask = None
        
        # Get initial dataset
        self.data, self.targets = [], []
        for i in range(len(self.base_dataset)):
            d, t = self.base_dataset.__getitem__(i)
            if t not in self.get_classes:
                continue
            self.data.append(d)
            self.targets.append(t)

        # Select known        
        self.targets = torch.tensor(self.targets)
        self.original_targets = copy.deepcopy(self.targets)
        
        is_known = torch.any(torch.stack([torch.eq(self.targets, aelem).logical_or_(torch.eq(self.targets, aelem)) for aelem in self.known_classes], dim=0), dim = 0)

        # Reorder all the known classes
        for new_label, label in enumerate(np.sort(self.known_classes)):
            self.targets[self.targets == label] = new_label
            
        # Set all the unknown as the last class
        self.targets[~is_known] = len(self.known_classes)
        
        # Get with specific indexes
        if isinstance(self.indexes, str) and self.indexes == 'all':
            pass
        else:
            self.data = itemgetter(*indexes)(self.data)
            self.targets = itemgetter(*indexes)(self.targets)
            self.original_targets = itemgetter(*indexes)(self.original_targets)
            
        # Remove unknown if return_only_known
        if self.return_only_known:
            self.data = list(compress(self.data, is_known))
            self.targets = list(compress(self.targets, is_known))
            self.original_targets = list(compress(self.original_targets, is_known))
        
    def reindex(self, new_index_mask):
        self.new_index_mask = new_index_mask
            
    def __getitem__(self, index: int) -> Tuple[Any, Any]:
        if self.new_index_mask is not None:
            index = self.new_index_mask[index]
            
        d, t = self.data[index], self.targets[index]
            
        if self.onehot:
            transform = onehot_bg(self.onehot_num_classes)
            t = transform(t)
        return d,t

    def __len__(self) -> int:
        if self.new_index_mask is not None:
            return len(self.new_index_mask)

        return len(self.targets)

    def _num_classes(self):
        return len(np.unique(self.targets))
    


from .base_datasets import *
from .OutlierDataset import OutlierDataset
from .DataWrapper import DataWrapper
from .datasets_config import (configure_division, get_train_test)
from .datasets_config_oneclass import (configure_oneclass_division)
from .datasets_outlier_config import (configure_division_outlier, get_train_test_outlier)

__all__ = [
    "base_datasets",
    "DataWrapper",
    "OutlierDataset",
    "configure_division",
    "configure_oneclass_division",
    "configure_division_outlier",
    "get_train_test",
    "get_train_test_outlier"
]



from .architectures import (fc_lower_stack, flat_fc_lower_stack, osrci_lower_stack, alexNet32_lower_stack,
                            flat_empty_lower_stack)

__all__ = [
    "fc_lower_stack",
    "flat_fc_lower_stack",
    "osrci_lower_stack",
    "alexNet32_lower_stack",
    "flat_empty_lower_stack"
]



import torch
from torch import nn
from torchmetrics import Accuracy, ConfusionMatrix
from .TSoftmax import TSoftmax
from scipy.stats import exponweib
import numpy as np

class Openmax(TSoftmax):
    """
    Implementation of Openmax method

    :type lower_stack: nn.Sequential
    :param lower_stack: Network architecture of lower_stack
    
    :type n_known: int
    :param n_known: Number of known classes
    
    :type epsilon: float
    :param epsilon: Threshold for prediction probability
    
    :type tail: int
    :param tail: Tail size for estimating the parameters of Weibull distribution
    
    :type alpha: int
    :param alpha: Alpha (number of most significant classes to revise)
    """
    
    def __init__(self, lower_stack=None, n_known=3, epsilon=.5, tail=20, alpha=2):
        super(Openmax, self).__init__(lower_stack=lower_stack,
                                      n_known=n_known,
                                      epsilon=epsilon)
        self.tail = tail
        self.alpha = alpha
        self.weibs = {}

    def train(self, dataloader, loss_fn, optimizer):
        size = len(dataloader.dataset)
        
        y_flat = []
        correct = []
        activations = []
        
        for batch, (X, y) in enumerate(dataloader):
            # Compute prediction and loss
            pred = self(X)
            loss = loss_fn(pred, y[:,:-1])
            
            # Backpropagation
            optimizer.zero_grad()
            loss.backward()
            optimizer.step()
            
            # Estimate correctly classified examples
            _pred_flat = pred.argmax(1)
            _y_flat = y.argmax(1)            
            _correct = torch.eq(_pred_flat, _y_flat)

            activations.append(pred)
            y_flat.append(_y_flat)
            correct.append(_correct)
        
            if self.verbose:
                if batch % 100 == 0:
                    loss, current = loss.item(), batch * len(X)
                    print(f"loss: {loss:>7f}  [{current:>5d}/{size:>5d}]")

        activations = torch.cat(activations)
        correct = torch.cat(correct)
        y_flat = torch.cat(y_flat)
        
        activations = activations[correct]
        y_flat = y_flat[correct]        
        
        # Establish mean activation vectors for correctly classified known classes        
        n_known_classes = activations.shape[1]
        self.weibs = {}
        
        for label in range(n_known_classes):
            correct_subset = activations[y_flat == label]
            if correct_subset.shape[0] > self.tail:
                mav = correct_subset.mean(0).view(1, -1)

                # Distance between each correctly classified training example and MAV for particular class is computed to obtain class specific distance distribution. 
                # Calculate distance between correct_subset and mav.            
                dist = torch.cdist(mav, correct_subset)
                
                # Leave tail
                dist = dist.sort()[0][:,-self.tail:]
                
                # Fit distribution
                proba = dist[0].detach().numpy()
                try:
                    self.weibs.update({label: exponweib.fit(proba, loc=0, scale=.2)})
                except:
                    print('[WEIB] Skipped', proba)
        
    def test(self, dataloader, loss_fn, conf=False): 
        inner_preds = []
        inner_preds_hp = []
        inner_preds_overall = []
        outer_preds = []
        
        inner_targets = []
        outer_targets = []
        overall_targets = []
        
        # Define metrics for open set detection
        try:
            inner_metric = Accuracy(task="multiclass", num_classes=self.n_known, average='macro')
            overall_metric = Accuracy(task="multiclass", num_classes=self.n_known+1, average='macro')
        except:
            inner_metric = None
            overall_metric = None
            
        outer_metric = Accuracy(task='multiclass', num_classes=2, average='macro')

        with torch.no_grad():
            for X, y in dataloader:
                # Get y flat and known mask
                y_flat = y.argmax(1)
                known_mask = ~(y_flat == self.n_known)
                
                # Calculate logits for full batch
                logits = self(X)
                slogit = logits.sum(1)
                
                # Rank activations for alpha trimming
                a_trimmer = 1 - torch.nn.functional.one_hot(torch.topk(logits, self.alpha, dim=1)[1],
                                                            logits.shape[1]).sum(1)
                                
                # Calculate weights
                w = np.ones(logits.shape).astype(float)
                for i in self.weibs:
                    rv = self.weibs[i]
                    w[:,i] -= exponweib.cdf(logits[:,i], *rv)
                
                w = torch.Tensor(w)
                try:
                    w[a_trimmer] = 1
                except:
                    print('[a_trimmer] Skipped', w, a_trimmer)
                
                # Establish weighted logits and unknown activation
                v_logits = w*logits
                plogit = v_logits.sum(1)
                unknown_activation = torch.sub(slogit, plogit)

                # Calculate corrected activation
                logits = torch.cat((v_logits, unknown_activation[:,None]), 1)          

                # Establish inner preds and inner y
                inner_pp = nn.Softmax(dim=1)(logits[:,:-1])
                inner_pred = inner_pp.argmax(1)[known_mask]
                inner_target = y_flat[known_mask]           
                             
                # Establish outer pred
                # Check if prediction (max support) is for KKC
                pred_known = (nn.Softmax(dim=1)(logits).argmax(1) != self.n_known).int() # 1 = KKC

                # Check if support is sufficient to keep the decision
                pred_sure = (nn.Softmax(dim=1)(logits).max(1).values > self.epsilon).int() # 1 = KKC

                outer_pred = pred_known * pred_sure
                outer_target = (y_flat != self.n_known).int()
                
                inner_pp_overall = nn.Softmax(dim=1)(logits)
                inner_pred_overall = inner_pp_overall.argmax(1)
                inner_pred_overall[outer_pred==0]=self.n_known
                inner_pred_hp= inner_pred_overall[known_mask]
                
                # Store predictions
                inner_preds.append(inner_pred)
                inner_preds_hp.append(inner_pred_hp)
                inner_preds_overall.append(inner_pred_overall)
                inner_targets.append(inner_target)
                overall_targets.append(y_flat)
                
                outer_preds.append(outer_pred)
                outer_targets.append(outer_target)
                
        inner_targets = torch.cat(inner_targets)
        overall_targets = torch.cat(overall_targets)
        inner_preds = torch.cat(inner_preds)
        inner_preds_hp = torch.cat(inner_preds_hp)
        inner_preds_overall = torch.cat(inner_preds_overall)
        
        outer_targets = torch.cat(outer_targets)
        outer_preds = torch.cat(outer_preds)
        
        # Calculate scores
        if inner_metric is not None:
            inner_score = inner_metric(inner_preds, inner_targets)
            inner_score_hp = overall_metric(inner_preds_hp, inner_targets)
            inner_score_overall = overall_metric(inner_preds_overall, overall_targets)
        else:
            inner_score = torch.nan
            inner_score_hp = torch.nan
            inner_score_overall = torch.nan
            
        outer_score = outer_metric(outer_preds, outer_targets)
        
        if conf==True:
            c_inner = ConfusionMatrix(task="multiclass", num_classes=self.n_known)(inner_preds, inner_targets)
            c_outer = ConfusionMatrix(task="binary")(outer_preds, outer_targets)
            c_hp = ConfusionMatrix(task="multiclass", num_classes=self.n_known+1)(inner_preds_hp, inner_targets)
            c_overall = ConfusionMatrix(task="multiclass", num_classes=self.n_known+1)(inner_preds_overall, overall_targets)
            return inner_score, outer_score, inner_score_hp, inner_score_overall, c_inner, c_outer, c_hp, c_overall
        
        if self.verbose:
            print('Inner metric : %.3f' % inner_score)
            print('Outer metric : %.3f' % outer_score)
            
        return inner_score, outer_score, inner_score_hp, inner_score_overall


    def predict(self, X): 
        with torch.no_grad():
            logits = self(X)
            slogit = logits.sum(1)
            
            # Rank activations for alpha trimming
            a_trimmer = 1 - torch.nn.functional.one_hot(torch.topk(logits, self.alpha, dim=1)[1],
                                                        logits.shape[1]).sum(1)
                            
            # Calculate weights
            w = np.ones(logits.shape).astype(float)
            for i in self.weibs:
                rv = self.weibs[i]
                w[:,i] -= exponweib.cdf(logits[:,i], *rv)
            
            w = torch.Tensor(w)
            try:
                w[a_trimmer] = 1
            except:
                print('[a_trimmer] Skipped', w, a_trimmer)
            
            # Establish weighted logits and unknown activation
            v_logits = w*logits
            plogit = v_logits.sum(1)
            unknown_activation = torch.sub(slogit, plogit)

            # Calculate corrected activation
            logits = torch.cat((v_logits, unknown_activation[:,None]), 1)          

            pred_known = (nn.Softmax(dim=1)(logits).argmax(1) != self.n_known).int() # 1 = KKC
            pred_sure = (nn.Softmax(dim=1)(logits).max(1).values > self.epsilon).int() # 1 = KKC
            outer_pred = pred_known * pred_sure
            
            inner_pred_overall = nn.Softmax(dim=1)(logits).argmax(1)
            inner_pred_overall[outer_pred==0]=self.n_known
            
            return inner_pred_overall


import numpy as np
from sklearn.datasets import make_classification
from torch.utils.data import Dataset
from typing import Any, Tuple

class Madelon_base(Dataset):
    def __init__(self, **kwargs):
        self.kwargs = kwargs
        self.data, self.targets = self._load_data()

    def _load_data(self):
        data, targets = make_classification(**self.kwargs)
        return data.astype(np.float32), targets

    def __getitem__(self, index: int) -> Tuple[Any, Any]:
        return self.data[index], self.targets[index]

    def __len__(self) -> int:
        return len(self.targets)
    
    def _n_classes(self) -> int:
        return len(np.unique(self.targets))



import torch
from torch import nn
from ..architectures.architectures import fc_lower_stack

class OSRModule(nn.Module):
    def __init__(self, n_known, lower_stack=None, verbose=False):
        super(OSRModule, self).__init__()
        self.n_known = n_known
        self.verbose = verbose
        
        if lower_stack is None:
            self.lower_stack = fc_lower_stack(1, 28, 64)
        else:
            self.lower_stack = lower_stack
        
    def forward(self, X):
        representation = self.lower_stack(X)

        logits = self.upper_stack(representation)
        logits = self._norm(logits)
              
        return logits
    
    def train(self, dataloader, loss_fn, optimizer):
        size = len(dataloader.dataset)
        
        for batch, (X, y) in enumerate(dataloader):
            # Compute prediction and loss
            pred = self(X)
            loss = loss_fn(pred, y[:,:-1])
            
            # Backpropagation
            optimizer.zero_grad()
            loss.backward()
            optimizer.step()
            
            if self.verbose:
                if batch % 100 == 0:
                    loss, current = loss.item(), batch * len(X)
                    print(f"loss: {loss:>7f}  [{current:>5d}/{size:>5d}]")                  

        
    def _norm(self, x):
        norm = torch.norm(x, p=2, dim=1)
        x = x / (norm.expand(1, -1).t() + .0001)
        return x


from os import listdir
from os.path import join
from typing import Any, Callable, List, Optional, Tuple
import glob 
import numpy as np

from PIL import Image

from torchvision.datasets import VisionDataset
from torchvision.datasets.utils import check_integrity, download_and_extract_archive


class Omniglot_base(VisionDataset):
    """
    `Omniglot <https://github.com/brendenlake/omniglot>`_ Dataset.
    
    Base implementation from torch library.
    """

    folder = "omniglot-py"
    download_url_prefix = "https://raw.githubusercontent.com/brendenlake/omniglot/master/python"
    zips_md5 = {
        "images_background": "68d2efa1b9178cc56df9314c21c6e718",
        "images_evaluation": "6b91aef0f799c5bb55b94e3f2daec811",
    }

    def __init__(
        self,
        root: str,
        background: bool = True,
        transform: Optional[Callable] = None,
        target_transform: Optional[Callable] = None,
        download: bool = True,
    ) -> None:
        super().__init__(join(root, self.folder), transform=transform, target_transform=target_transform)
        self.background = background

        if download:
            self.download()
            
        if not self._check_integrity():
            raise RuntimeError("Dataset not found or corrupted. You can use download=True to download it")

        self.data, self.targets = self._load_data()
        

    def _load_data(self):
        self.target_folder = join(self.root, self._get_target_folder())
        self._alphabets = listdir(self.target_folder)
        try:
            self._alphabets.remove('.DS_Store')
        except:
            pass

        self._characters: List[str] = sum(
            ([join(a, c) for c in listdir(join(self.target_folder, a))] for a in self._alphabets), []
        )

        self._character_images = [
            [(image, idx) for image in glob.glob(join(self.target_folder, character, '*.png'))]
            
            for idx, character in enumerate(self._characters)
        ]
        
        self._flat_character_images: List[Tuple[str, int]] = sum(self._character_images, [])
        
        data = [Image.open(i, mode="r").convert("L") for i, _ in self._flat_character_images]
        targets = [i for _,i in self._flat_character_images]
        
        return data, targets
    
    def __len__(self) -> int:
        return len(self._flat_character_images)

    def __getitem__(self, index: int) -> Tuple[Any, Any]:
        image, target = self.data[index], self.targets[index]

        if self.transform:
            image = self.transform(image)

        if self.target_transform:
            target = self.target_transform(target)

        return image, target

    def _check_integrity(self) -> bool:
        zip_filename = self._get_target_folder()
        if not check_integrity(join(self.root, zip_filename + ".zip"), self.zips_md5[zip_filename]):
            return False
        return True

    def download(self) -> None:
        if self._check_integrity():
            print("Files already downloaded and verified")
            return

        filename = self._get_target_folder()
        zip_filename = filename + ".zip"
        url = self.download_url_prefix + "/" + zip_filename
        download_and_extract_archive(url, self.root, filename=zip_filename, md5=self.zips_md5[filename])

    def _get_target_folder(self) -> str:
        return "images_background" if self.background else "images_evaluation"
    
    def _n_classes(self) -> int:
        return len(np.unique(self.targets))


from .Openmax import (Openmax)
from .TSoftmax import (TSoftmax)
from .NoiseSoftmax import (NoiseSoftmax)
from .OverlaySoftmax import (OverlaySoftmax)
from .base import (OSRModule)
from .GSL import (GSL)

__all__ = [
    "OSRModule",
    "Openmax",
    "TSoftmax",
    "NoiseSoftmax",
    "OverlaySoftmax",
    "GSL"
]



import torch
from PIL import ImageOps
from torchvision.transforms import Lambda

def onehot_bg(n_classes):
    """
    One-hot encoder for classification labels
    """
    return Lambda(lambda y: torch.zeros(n_classes, dtype=torch.float).scatter_(0, y.clone().detach(), value=1))

def inverse_transform():
    """Inverse image values"""
    return Lambda(lambda img: 1-img)

def grayscale_transform():
    """Return mean on color channels"""    
    return Lambda(lambda img: ImageOps.grayscale(img))

def get_openmax_epsilon(n_known):
    """Optimized epsilon value for Openmax and given number of KKC"""
    return {9:.1, 8:.1, 7:.1, 6:.1,
            5:.2, 4:.4, 3:.5, 2:.6}[n_known]
        
def get_softmax_epsilon(n_known):
    """Optimized epsilon value for Thresholded Softmax and given number of KKC"""
    return {10: .04, 9:.05, 8:.1, 7:.2, 6:.25, 
            5:.35, 4:.48, 3:.6, 2:.8}[n_known]